"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.AppFactory = void 0;
const core_1 = require("@my-devkit/core");
const app_1 = require("firebase-admin/app");
const auth_1 = require("firebase-admin/auth");
const firestore_1 = require("firebase-admin/firestore");
const params_1 = require("firebase-functions/params");
const v1_1 = require("firebase-functions/v1");
const v2_1 = require("firebase-functions/v2");
const firestore_2 = require("firebase-functions/v2/firestore");
const https_1 = require("firebase-functions/v2/https");
const bus_1 = require("./bus");
const constants_1 = require("./constants");
const context_1 = require("./context");
const execution_mode_enum_1 = require("./execution-mode-enum");
const firestore_client_1 = require("./firestore-client");
const reflect_1 = require("./reflect");
const server_1 = require("./server");
class AppFactory {
    static isActive(name) {
        const functionName = process.env.FUNCTION_NAME || process.env.K_SERVICE;
        return this.isEmulator || !functionName || functionName.toLowerCase() === name.toLowerCase();
    }
    static initializeApp(options) {
        this.instanceId = (0, core_1.guid)();
        this.options = options;
        context_1.Context.init();
        this.registerLogger();
        this.publicPaths = options.publicPaths || [];
        const app = (0, app_1.initializeApp)();
        this.firestore = (0, firestore_1.getFirestore)(app);
        this.firestore.settings({});
        this.apiKey = (0, params_1.defineSecret)('GOOGLE_GENAI_API_KEY');
        this.auth = (0, auth_1.getAuth)(app);
        // THIS PART SHOULD BE REMOVED ONCE COMPLETELY UPGRADED TO V2
        const runtimeOptions = {
            timeoutSeconds: options.timeoutSeconds,
            memory: '1GB'
        };
        this.functionInRegion = (0, v1_1.region)(this.isEmulator ? 'functions' : options.region).runWith(runtimeOptions);
    }
    static createHttpsFunction(name, module) {
        if (!this.isActive(name)) {
            return null;
        }
        core_1.Logger.info(`Cold start Http function ${name}`);
        const bus = new bus_1.Bus();
        const server = new server_1.Server({
            bus,
            firebaseAuth: this.auth,
            firestore: this.firestore,
            publicPaths: this.publicPaths
        });
        this.registerControllers(server, module);
        this.registerEventPersisterHandler(bus);
        this.registerHandlers(bus, module);
        return (0, https_1.onRequest)({
            memory: this.options.memory,
            timeoutSeconds: this.options.timeoutSeconds,
            region: this.options.region
        }, server.app);
    }
    static createOnCallGenkit(name, flow) {
        if (!this.isActive(name))
            return null;
        core_1.Logger.info(`Cold start Genkit callable function ${name}`);
        return (0, https_1.onCallGenkit)({
            secrets: [this.apiKey],
            region: this.options.region,
            cors: this.options.cors
        }, flow);
    }
    static createScheduledJob(name, module, schedule, timeZone, commandType) {
        if (!this.isActive(name))
            return null;
        core_1.Logger.info(`Cold start scheduled job ${name}`);
        const bus = new bus_1.Bus();
        this.registerEventPersisterHandler(bus);
        this.registerHandlers(bus, module);
        // V2 gets errors on deployed environment (PERMISSION_DENIED...)
        // return onSchedule({
        //     schedule,
        //     timeZone,
        //     memory: this.options.memory,
        //     timeoutSeconds: this.options.timeoutSeconds,
        //     region: this.options.region
        // }, this.getFunctionCommandHandler(bus, commandType));
        return this.functionInRegion
            .pubsub
            .schedule(schedule)
            .timeZone(timeZone)
            .onRun(this.getFunctionCommandHandler(bus, commandType));
    }
    static createEventListener(name, module, mode = execution_mode_enum_1.ExecutionMode.Asynchronous) {
        if (!this.isActive(name))
            return null;
        core_1.Logger.info(`Cold start Event Listener ${mode}`);
        const bus = new bus_1.Bus();
        this.registerEventPersisterHandler(bus);
        this.registerHandlers(bus, module);
        return (0, firestore_2.onDocumentCreated)({
            document: 'events/{eventId}',
            memory: this.options.memory,
            timeoutSeconds: this.options.timeoutSeconds,
            region: this.options.region,
        }, this.getFunctionEventHandler(bus, execution_mode_enum_1.ExecutionMode.Asynchronous));
    }
    static registerEventPersisterHandler(bus) {
        bus.onEachEvent((event) => __awaiter(this, void 0, void 0, function* () {
            context_1.Context.resolve(firestore_client_1.FirestoreClient).createDocument(event);
        }), execution_mode_enum_1.ExecutionMode.Synchronous);
    }
    static registerControllers(server, module) {
        const controllers = Reflect.getMetadata(constants_1.MODULE_METADATA.CONTROLLERS, module) || [];
        for (const controller of controllers) {
            this.registerController(server, controller);
        }
        const imports = Reflect.getMetadata(constants_1.MODULE_METADATA.IMPORTS, module) || [];
        for (const importedModule of imports) {
            this.registerControllers(server, importedModule);
        }
    }
    static registerController(server, controller) {
        const router = new server_1.Server.Router();
        const config = (0, reflect_1.reflect)(controller).getControllerConfiguration();
        for (const route of config.routes) {
            router.registerRoute(route.requestMethod, `/${route.path}`, (request, response) => __awaiter(this, void 0, void 0, function* () {
                core_1.Logger.info(`${route.requestMethod}/${route.path}: executing handler method ${controller.name}.${route.methodName}`);
                const controllerInstance = context_1.Context.resolve(controller);
                const args = route.argumentInjectors.map(injector => injector(request));
                const result = yield controllerInstance[route.methodName](...args);
                yield context_1.Context.commit();
                context_1.Context.destroy();
                response.status(server_1.Server.statusCodes.OK).send({ result });
            }));
        }
        router.registerOnServer(`/${config.basePath}`, server);
    }
    static registerHandlers(bus, module) {
        const handlers = Reflect.getMetadata(constants_1.MODULE_METADATA.HANDLERS, module) || [];
        for (const handler of handlers) {
            const handlerConfiguration = (0, reflect_1.reflect)(handler).getHandlerConfiguration();
            const instance = new handler();
            for (const commandHandler of Array.from(handlerConfiguration.commandHandlers.values())) {
                bus.onCommand(commandHandler.command, (command) => {
                    const startDate = new Date();
                    core_1.Logger.info(`${handler.name}.${commandHandler.methodName} handle ${command._type}`);
                    const result = instance[commandHandler.methodName](command);
                    core_1.Logger.infoWithElapesedTime(`${command._type} handled by ${handler.name}.${commandHandler.methodName}`, startDate);
                    return result;
                });
            }
            for (const eventHandler of Array.from(handlerConfiguration.eventHandlers.values())) {
                bus.onEvent(eventHandler.event, (event) => __awaiter(this, void 0, void 0, function* () {
                    const startDate = new Date();
                    core_1.Logger.info(`${handler.name}.${eventHandler.methodName} handle ${event._type}`);
                    yield instance[eventHandler.methodName](event);
                    core_1.Logger.infoWithElapesedTime(`${event._type} handled by ${handler.name}.${eventHandler.methodName}`, startDate);
                }), eventHandler.executionMode);
            }
        }
        const imports = Reflect.getMetadata(constants_1.MODULE_METADATA.IMPORTS, module) || [];
        for (const importedModule of imports) {
            this.registerHandlers(bus, importedModule);
        }
    }
    static registerLogger() {
        if (this.isEmulator) {
            core_1.Logger.registerLogger(new core_1.Logger.ConsoleImplementation());
            return;
        }
        const completeMetaData = (data) => {
            var _a, _b;
            return Object.assign({ instanceId: this.instanceId, correlationId: context_1.Context.correlationId, eventId: context_1.Context.eventId, userId: (_a = context_1.Context.user) === null || _a === void 0 ? void 0 : _a.uid, userEmail: (_b = context_1.Context.user) === null || _b === void 0 ? void 0 : _b.email }, data);
        };
        core_1.Logger.registerLogger({
            debug: (message, data) => {
                v2_1.logger.debug(message, completeMetaData(data));
            },
            info: (message, data) => {
                v2_1.logger.info(message, completeMetaData(data));
            },
            warn: (message, data) => {
                v2_1.logger.warn(message, completeMetaData(data));
            },
            error: (message, data) => {
                v2_1.logger.error(message, completeMetaData(data));
            }
        });
    }
    static getFunctionCommandHandler(bus, commandType) {
        return () => __awaiter(this, void 0, void 0, function* () {
            const startDate = new Date();
            const command = new commandType();
            context_1.Context.create(bus);
            context_1.Context.correlationId = (0, core_1.guid)();
            try {
                core_1.Logger.info(`${'-'.repeat(5)} handling ${command._type} ${'-'.repeat(5)}`);
                yield bus.send(command);
                yield context_1.Context.commit();
                core_1.Logger.infoWithElapesedTime(`${'-'.repeat(5)} ${command._type} handled successfully ${'-'.repeat(5)}`, startDate);
            }
            catch (error) {
                core_1.Logger.error(`${'-'.repeat(5)} ${command._type} couldn't be handled correctly ${'-'.repeat(5)}`, { error });
            }
            finally {
                context_1.Context.destroy();
            }
        });
    }
    static getFunctionEventHandler(bus, mode) {
        return (snapshot) => __awaiter(this, void 0, void 0, function* () {
            const startDate = new Date();
            const unserializedEvent = snapshot.data.data();
            const event = (0, core_1.deserialize)(unserializedEvent);
            context_1.Context.create(bus);
            context_1.Context.correlationId = event._correlationId;
            context_1.Context.eventId = event.id;
            if (unserializedEvent['_restored'] === true) {
                core_1.Logger.debug(`Event ${event._type} is ignored as it is restored from backup.`);
                return;
            }
            try {
                core_1.Logger.info(`${'-'.repeat(5)} handling ${event._type} ${'-'.repeat(5)}`);
                yield bus.publish(event, mode);
                yield context_1.Context.commit();
                core_1.Logger.infoWithElapesedTime(`${'-'.repeat(5)} ${event._type} handled successfully ${'-'.repeat(5)}`, startDate);
            }
            catch (error) {
                core_1.Logger.error(`${'-'.repeat(5)} ${event._type} couldn't be handled correctly ${'-'.repeat(5)}`, { error });
            }
            finally {
                context_1.Context.destroy();
            }
        });
    }
    static get isEmulator() {
        return !!process.env.FUNCTIONS_EMULATOR;
    }
}
exports.AppFactory = AppFactory;
//# sourceMappingURL=app-factory.js.map