import { Command, deserialize, Event, guid, Logger } from '@my-devkit/core';
import { initializeApp } from 'firebase-admin/app';
import { Auth, getAuth } from 'firebase-admin/auth';
import { Firestore, getFirestore, QueryDocumentSnapshot } from 'firebase-admin/firestore';
import { defineSecret } from 'firebase-functions/params';
import { FunctionBuilder, region } from 'firebase-functions/v1';
import { logger, MemoryOption, SupportedRegion } from 'firebase-functions/v2';

import { SecretParam } from 'firebase-functions/lib/params/types';
import { RuntimeOptions } from 'firebase-functions/lib/v1/function-configuration';
import { FirestoreEvent, onDocumentCreated } from 'firebase-functions/v2/firestore';
import { HttpsFunction, onCallGenkit, onRequest } from 'firebase-functions/v2/https';
import { Bus } from './bus';
import { MODULE_METADATA } from './constants';
import { Context } from './context';
import { ExecutionMode } from './execution-mode-enum';
import { FirestoreClient } from './firestore-client';
import { Type } from './interfaces';
import { reflect } from './reflect';
import { Server } from './server';

export class AppFactory {
    private static functionInRegion: FunctionBuilder;
    private static firestore: Firestore;
    private static auth: Auth;
    private static publicPaths: string[];
    private static instanceId: string;
    private static options: AppFactory.Options;
    private static apiKey: SecretParam;

    public static isActive(name: string): boolean {
        const functionName = process.env.FUNCTION_NAME || process.env.K_SERVICE;
        return this.isEmulator || !functionName || functionName.toLowerCase() === name.toLowerCase();
    }

    public static initializeApp(options: AppFactory.Options): void {
        this.instanceId = guid();
        this.options = options;

        Context.init();
        this.registerLogger();

        this.publicPaths = options.publicPaths || [];

        const app = initializeApp();
        this.firestore = getFirestore(app);
        this.firestore.settings({});

        this.apiKey = defineSecret('GOOGLE_GENAI_API_KEY');

        this.auth = getAuth(app);

        // THIS PART SHOULD BE REMOVED ONCE COMPLETELY UPGRADED TO V2
        const runtimeOptions: RuntimeOptions = {
            timeoutSeconds: options.timeoutSeconds,
            memory: '1GB'
        };

        this.functionInRegion = region(this.isEmulator ? 'functions' : options.region).runWith(runtimeOptions);
    }

    public static createHttpsFunction(name: string, module: Type): HttpsFunction {
        if (!this.isActive(name)) {
            return null;
        }

        Logger.info(`Cold start Http function ${name}`);

        const bus = new Bus();

        const server = new Server({
            bus,
            firebaseAuth: this.auth,
            firestore: this.firestore,
            publicPaths: this.publicPaths
        });

        this.registerControllers(server, module);

        this.registerEventPersisterHandler(bus);

        this.registerHandlers(bus, module);

        return onRequest({
            memory: this.options.memory,
            timeoutSeconds: this.options.timeoutSeconds,
            region: this.options.region
        }, server.app);
    }

    public static createOnCallGenkit(name: string, flow: any): CallableFunction {
        if (!this.isActive(name))
            return null;

        Logger.info(`Cold start Genkit callable function ${name}`);

        return onCallGenkit({
            secrets: [this.apiKey],
            region: this.options.region,
            cors: this.options.cors
        }, flow);
    }

    public static createScheduledJob<C extends Command>(
        name: string,
        module: Type,
        schedule: string,
        timeZone: string,
        commandType: new () => C
    ) {
        if (!this.isActive(name))
            return null;

        Logger.info(`Cold start scheduled job ${name}`);

        const bus = new Bus();

        this.registerEventPersisterHandler(bus);

        this.registerHandlers(bus, module);

        // V2 gets errors on deployed environment (PERMISSION_DENIED...)
        // return onSchedule({
        //     schedule,
        //     timeZone,
        //     memory: this.options.memory,
        //     timeoutSeconds: this.options.timeoutSeconds,
        //     region: this.options.region
        // }, this.getFunctionCommandHandler(bus, commandType));

        return this.functionInRegion
            .pubsub
            .schedule(schedule)
            .timeZone(timeZone)
            .onRun(this.getFunctionCommandHandler(bus, commandType));
    }

    public static createEventListener(name: string, module: Type, mode: ExecutionMode = ExecutionMode.Asynchronous) {
        if (!this.isActive(name))
            return null;

        Logger.info(`Cold start Event Listener ${mode}`);

        const bus = new Bus();

        this.registerEventPersisterHandler(bus);

        this.registerHandlers(bus, module);

        return onDocumentCreated({
            document: 'events/{eventId}',
            memory: this.options.memory,
            timeoutSeconds: this.options.timeoutSeconds,
            region: this.options.region,
        }, this.getFunctionEventHandler(bus, ExecutionMode.Asynchronous));
    }

    private static registerEventPersisterHandler(bus: Bus): void {
        bus.onEachEvent(async (event: Event): Promise<void> => {
            Context.resolve(FirestoreClient).createDocument(event);
        }, ExecutionMode.Synchronous);
    }

    private static registerControllers(server: Server, module: Type): void {
        const controllers = Reflect.getMetadata(MODULE_METADATA.CONTROLLERS, module) || [];
        for (const controller of controllers) {
            this.registerController(server, controller);
        }

        const imports: any[] = Reflect.getMetadata(MODULE_METADATA.IMPORTS, module) || [];
        for (const importedModule of imports) {
            this.registerControllers(server, importedModule);
        }
    }

    private static registerController(server: Server, controller: Type): void {
        const router = new Server.Router();

        const config = reflect(controller).getControllerConfiguration();

        for (const route of config.routes) {
            router.registerRoute(route.requestMethod, `/${route.path}`, async (request: Server.Request, response: Server.Response): Promise<void> => {
                Logger.info(`${route.requestMethod}/${route.path}: executing handler method ${controller.name}.${route.methodName}`);
                const controllerInstance = Context.resolve(controller);
                const args = route.argumentInjectors.map(injector => injector(request));
                const result = await controllerInstance[route.methodName](...args);

                await Context.commit();

                Context.destroy();
                response.status(Server.statusCodes.OK).send({ result });
            });
        }

        router.registerOnServer(`/${config.basePath}`, server);
    }

    private static registerHandlers(bus: Bus, module: Type): void {
        const handlers = Reflect.getMetadata(MODULE_METADATA.HANDLERS, module) || [];
        for (const handler of handlers) {
            const handlerConfiguration = reflect(handler).getHandlerConfiguration();

            const instance = new handler();
            for (const commandHandler of Array.from(handlerConfiguration.commandHandlers.values())) {
                bus.onCommand(commandHandler.command, (command) => {
                    const startDate = new Date();
                    Logger.info(`${handler.name}.${commandHandler.methodName} handle ${command._type}`);
                    const result = instance[commandHandler.methodName](command);
                    Logger.infoWithElapesedTime(`${command._type} handled by ${handler.name}.${commandHandler.methodName}`, startDate);
                    return result;
                });
            }

            for (const eventHandler of Array.from(handlerConfiguration.eventHandlers.values())) {
                bus.onEvent(eventHandler.event, async (event) => {
                    const startDate = new Date();
                    Logger.info(`${handler.name}.${eventHandler.methodName} handle ${event._type}`);
                    await instance[eventHandler.methodName](event);
                    Logger.infoWithElapesedTime(`${event._type} handled by ${handler.name}.${eventHandler.methodName}`, startDate);
                }, eventHandler.executionMode);
            }
        }

        const imports: any[] = Reflect.getMetadata(MODULE_METADATA.IMPORTS, module) || [];
        for (const importedModule of imports) {
            this.registerHandlers(bus, importedModule);
        }
    }

    private static registerLogger(): void {
        if (this.isEmulator) {
            Logger.registerLogger(new Logger.ConsoleImplementation());
            return;
        }

        const completeMetaData = (data: Logger.LogMetaData): Logger.LogMetaData => {
            return {
                instanceId: this.instanceId,
                correlationId: Context.correlationId,
                eventId: Context.eventId,
                userId: Context.user?.uid,
                userEmail: Context.user?.email,
                ...data
            };
        };

        Logger.registerLogger({
            debug: (message: string, data: Logger.LogMetaData): void => {
                logger.debug(message, completeMetaData(data));
            },
            info: (message: string, data: Logger.LogMetaData): void => {
                logger.info(message, completeMetaData(data));
            },
            warn: (message: string, data: Logger.LogMetaData): void => {
                logger.warn(message, completeMetaData(data));
            },
            error: (message: string, data: Logger.LogMetaData): void => {
                logger.error(message, completeMetaData(data));
            }
        });
    }

    public static getFunctionCommandHandler<C extends Command>(bus: Bus, commandType: new () => C): () => any {
        return async (): Promise<void> => {
            const startDate = new Date();

            const command = new commandType();

            Context.create(bus);
            Context.correlationId = guid();

            try {
                Logger.info(`${'-'.repeat(5)} handling ${command._type} ${'-'.repeat(5)}`);

                await bus.send(command);

                await Context.commit();

                Logger.infoWithElapesedTime(`${'-'.repeat(5)} ${command._type} handled successfully ${'-'.repeat(5)}`, startDate);
            } catch (error) {
                Logger.error(`${'-'.repeat(5)} ${command._type} couldn't be handled correctly ${'-'.repeat(5)}`, { error });
            } finally {
                Context.destroy();
            }
        };
    }

    public static getFunctionEventHandler(bus: Bus, mode: ExecutionMode): (snapshot: FirestoreEvent<QueryDocumentSnapshot>) => any {
        return async (snapshot: FirestoreEvent<QueryDocumentSnapshot>): Promise<void> => {
            const startDate = new Date();
            const unserializedEvent = snapshot.data.data();

            const event = deserialize<Event>(unserializedEvent);

            Context.create(bus);
            Context.correlationId = event._correlationId;
            Context.eventId = event.id;

            if (unserializedEvent['_restored'] === true) {
                Logger.debug(`Event ${event._type} is ignored as it is restored from backup.`);
                return;
            }

            try {
                Logger.info(`${'-'.repeat(5)} handling ${event._type} ${'-'.repeat(5)}`);

                await bus.publish(event, mode);

                await Context.commit();

                Logger.infoWithElapesedTime(`${'-'.repeat(5)} ${event._type} handled successfully ${'-'.repeat(5)}`, startDate);
            } catch (error) {
                Logger.error(`${'-'.repeat(5)} ${event._type} couldn't be handled correctly ${'-'.repeat(5)}`, { error });
            } finally {
                Context.destroy();
            }
        };
    }

    public static get isEmulator(): boolean {
        return !!process.env.FUNCTIONS_EMULATOR;
    }
}

export namespace AppFactory {
    export interface Options {
        region: SupportedRegion;
        timeoutSeconds: number;
        memory: MemoryOption;
        cors: string[] | boolean;
        publicPaths?: string[]
    }
}
