import { Command, DomainError, Event, Logger, serialize, Validators } from '@my-devkit/core';

import { ExecutionMode } from './execution-mode-enum';

export class Bus {
    private commandHandlers = new Map<string, Bus.Handler<Command, any>>();
    private eventHandlers = new Map<string, Bus.Handler<Event, void>[]>();

    public async send<T, C extends Command>(command: C): Promise<T> {
        Logger.info(`Bus.send ${command._type}`, { command: serialize(command) });

        const validationErrors = Validators.validate(command);
        if (validationErrors.length > 0) {
            throw new DomainError(validationErrors[0].getErrorMessage());
        }

        const handler: Bus.Handler<C, T> = this.commandHandlers.get(command._type);
        if (!handler) {
            throw new Error(`No command handler for ${command._type}`);
        }
        Logger.info(`Run command handler ${handler.name}`);

        return handler(command);
    }

    public async publish<E extends Event>(event: E, mode: ExecutionMode = ExecutionMode.Synchronous): Promise<void> {
        Logger.info(`Bus.publish ${event._type} `, { event: serialize(event) });

        const handlers = [...this.getEventHandlers(event._type, mode), ...this.getEventHandlers('**', mode)];
        for (const handler of handlers) {
            Logger.info(`Run event handler ${handler.name}`);
            await handler(event);
        }
    }

    public onCommand<C extends Command, R>(commandType: new () => C, handler: Bus.Handler<C, R>): void {
        const commandName = commandType.constructor.name;
        this.commandHandlers.set(commandName, handler);
    }

    public onEvent<E extends Event>(eventType: new () => E, handler: Bus.Handler<E, void>, mode: ExecutionMode): void {
        this.registerEventHandler(eventType.constructor.name, mode, handler);
    }

    public onEachEvent<E extends Event>(handler: Bus.Handler<E, void>, mode: ExecutionMode): void {
        this.registerEventHandler('**', mode, handler);
    }

    private getDictionaryKey(eventType: string, mode: ExecutionMode): string {
        return `${eventType}_${mode.toString()}`;
    }

    private registerEventHandler<E extends Event>(type: string, mode: ExecutionMode, handler: Bus.Handler<E, void>): void {
        const key = this.getDictionaryKey(type, mode);

        if (!this.eventHandlers.has(key)) {
            this.eventHandlers.set(key, []);
        }
        this.eventHandlers.get(key).push(handler);
    }

    private getEventHandlers(type: string, mode: ExecutionMode) {
        const key = this.getDictionaryKey(type, mode);
        return this.eventHandlers.get(key) || [];
    }
}

export namespace Bus {
    export type Handler<T, R> = (arg: T) => Promise<R>;
}
