import { Logger } from '@my-devkit/core';
import { AsyncHook, createHook, executionAsyncId } from 'async_hooks';

import { Bus } from './bus';
import { Injector } from './injector';
import { ITransactionalClient, Type } from './interfaces';

interface ContextStore {
    injector: Injector;
    bus: Bus;
    correlationId?: string;
    eventId?: string;
    user?: UserRecord;
}

interface UserRecord {
    uid: string;
    email?: string;
    displayName?: string;
    photoURL?: string;
    disabled: boolean;
}

export class Context {
    private static store = new Map<number, ContextStore>();
    private static asyncHook: AsyncHook;

    public static init(): void {
        this.asyncHook = createHook({
            init: (asyncId: number, _type: string, triggerAsyncId: number) => {
                if (this.store.has(triggerAsyncId)) {
                    this.store.set(asyncId, this.store.get(triggerAsyncId))
                }
            },
            destroy: (asyncId: number) => {
                this.store.delete(asyncId);
            }
        });
        this.asyncHook.enable();
    }

    public static create(bus: Bus): void {
        this.store.set(executionAsyncId(), { injector: new Injector(), bus });
    }

    public static resolve<T>(type: Type<T>): T {
        if (!this.store.has(executionAsyncId())) {
            throw new Error(`Create context before resolving an injection!`);
        }
        return this.store.get(executionAsyncId()).injector.resolve(type);
    }

    public static getTransactionalClients(): ITransactionalClient[] {
        if (!this.store.has(executionAsyncId())) {
            throw new Error(`Create context before getting transactional clients!`);
        }
        return this.store.get(executionAsyncId()).injector.getTransactionalClients();
    }

    public static get bus(): Bus {
        return this.store.get(executionAsyncId())?.bus;
    }

    public static set correlationId(correlationId: string) {
        if (!this.store.has(executionAsyncId())) {
            throw new Error(`Create context before setting correlationId!`);
        }
        this.store.get(executionAsyncId()).correlationId = correlationId;
    }

    public static get correlationId(): string {
        return this.store.get(executionAsyncId())?.correlationId;
    }

    public static set eventId(eventId: string) {
        if (!this.store.has(executionAsyncId())) {
            throw new Error(`Create context before setting eventId!`);
        }
        this.store.get(executionAsyncId()).eventId = eventId;
    }

    public static get eventId(): string {
        return this.store.get(executionAsyncId())?.eventId;
    }

    public static set user(user: UserRecord) {
        if (!this.store.has(executionAsyncId())) {
            throw new Error(`Create context before setting user!`);
        }
        this.store.get(executionAsyncId()).user = user;
    }

    public static get user(): UserRecord {
        return this.store.get(executionAsyncId())?.user;
    }

    public static destroy(): void {
        this.store.delete(executionAsyncId());
    }

    public static size(): number {
        return this.store.size;
    }

    public static logRemainingKeys(): void {
        Logger.info(Array.from(this.store.keys()).join(' '));
    }

    public static async commit(): Promise<void> {
        Logger.info(`Commiting transactions...`);

        for (const client of Context.getTransactionalClients()) {
            await client.save();
        }
    }
}
