import { Command } from '@my-devkit/core';
import { expect } from 'chai';

import { reflect } from '../../reflect';
import { CommandHandler } from './command-handler';

describe('Given a CommandHandler decorator', () => {

    describe('When decorator is called', () => {

        describe('And decorated method has no parameter', () => {
            it('Then we should retrieve information via reflect', async () => {

                try {
                    class TestClass {
                        @CommandHandler()
                        public create(): string {
                            return `create`;
                        }
                    }

                    reflect(TestClass).getHandlerConfiguration();
                } catch (error) {
                    expect(error.message).equal('TestClass.create should have one argument!');
                }
            });
        });

        describe('And decorated method has more than 1 parameter', () => {
            it('Then we should retrieve information via reflect', async () => {

                try {
                    class TestClass {
                        @CommandHandler()
                        public create(param1: string, param2: string): string {
                            return `create ${param1} / ${param2}`;
                        }
                    }

                    reflect(TestClass).getHandlerConfiguration();
                } catch (error) {
                    expect(error.message).equal('TestClass.create should have only one argument!');
                }
            });
        });

        describe('And decorated method parameter is not a command', () => {
            it('Then we should retrieve information via reflect', async () => {

                try {
                    class FakeCommand {
                        test: string;
                    }

                    class TestClass {
                        @CommandHandler()
                        public create(fakeCommand: FakeCommand): string {
                            return `create ${fakeCommand}`;
                        }
                    }

                    reflect(TestClass).getHandlerConfiguration();
                } catch (error) {
                    expect(error.message).equal('TestClass.create should have a command as argument!');
                }
            });
        });

        describe('And handler is correctly implemented', () => {

            class CreateCommand extends Command {
            }

            class UpdateCommand extends Command {
            }

            class TestClass {
                @CommandHandler()
                public create(command: CreateCommand): string {
                    return `create: ${command}`;
                }

                @CommandHandler()
                public update(command: UpdateCommand): string {
                    return `update: ${command}`;
                }
            }
            it('Then we should retrieve information via reflect', async () => {
                const commandHandlers = Array.from(reflect(TestClass).getHandlerConfiguration().commandHandlers.values());

                expect(commandHandlers.length).equal(2, 'Wrong route count detected');

                expect(commandHandlers[0].methodName).equal('create');
                expect(commandHandlers[0].command.constructor.name).equal('CreateCommand');

                expect(commandHandlers[1].methodName).equal('update');
                expect(commandHandlers[1].command.constructor.name).equal('UpdateCommand');

            });
        });
    });
});
