import { Event } from '@my-devkit/core';
import { expect } from 'chai';

import { ExecutionMode } from '../../execution-mode-enum';
import { reflect } from '../../reflect';
import { EventHandler } from './event-handler';

describe('Given a EventHandler decorator', () => {

    describe('When decorator is called', () => {

        describe('And decorated method has no parameter', () => {
            it('Then we should retrieve information via reflect', async () => {

                try {
                    class TestClass {
                        @EventHandler(ExecutionMode.Asynchronous)
                        public created(): string {
                            return `created`;
                        }
                    }

                    reflect(TestClass).getHandlerConfiguration();
                } catch (error) {
                    expect(error.message).equal('TestClass.created should have one argument!');
                }
            });
        });

        describe('And decorated method has more than 1 parameter', () => {
            it('Then we should retrieve information via reflect', async () => {

                try {
                    class TestClass {
                        @EventHandler(ExecutionMode.Asynchronous)
                        public created(param1: string, param2: string): string {
                            return `created ${param1} / ${param2}`;
                        }
                    }

                    reflect(TestClass).getHandlerConfiguration();
                } catch (error) {
                    expect(error.message).equal('TestClass.created should have only one argument!');
                }
            });
        });

        describe('And decorated method parameter is not a command', () => {
            it('Then we should retrieve information via reflect', async () => {

                try {
                    class FakeEvent {
                        test: string;
                    }

                    class TestClass {
                        @EventHandler(ExecutionMode.Asynchronous)
                        public created(fakeEvent: FakeEvent): string {
                            return `created ${fakeEvent}`;
                        }
                    }

                    reflect(TestClass).getHandlerConfiguration();
                } catch (error) {
                    expect(error.message).equal('TestClass.created should have an event as argument!');
                }
            });
        });

        describe('And handler is correctly implemented', () => {

            class CreatedEvent extends Event {
            }

            class UpdatedEvent extends Event {
            }

            class TestClass {
                @EventHandler(ExecutionMode.Asynchronous)
                public created(event: CreatedEvent): string {
                    return `created: ${event}`;
                }

                @EventHandler(ExecutionMode.Synchronous)
                public updated(event: UpdatedEvent): string {
                    return `updated: ${event}`;
                }
            }
            it('Then we should retrieve information via reflect', async () => {
                const eventHandlers = Array.from(reflect(TestClass).getHandlerConfiguration().eventHandlers.values());

                expect(eventHandlers.length).equal(2, 'Wrong route count detected');

                expect(eventHandlers[0].methodName).equal('created');
                expect(eventHandlers[0].executionMode).equal(ExecutionMode.Asynchronous);
                expect(eventHandlers[0].event.constructor.name).equal('CreatedEvent');

                expect(eventHandlers[1].methodName).equal('updated');
                expect(eventHandlers[1].executionMode).equal(ExecutionMode.Synchronous);
                expect(eventHandlers[1].event.constructor.name).equal('UpdatedEvent');

            });
        });
    });
});
