import { Logger } from '@my-devkit/core';

import { Aggregate } from './aggregate';
import { Context } from './context';
import { ExecutionMode } from './execution-mode-enum';
import { FirestoreClient } from './firestore-client';

export class HandlerHelper {
    public static async create<A extends Aggregate, R>(aggregate: A, callback: (a: A) => Promise<R>): Promise<R> {
        const result = await callback(aggregate);

        this.client.createDocument(aggregate);
        await this.publishEvents(aggregate);

        return result;
    }

    public static async update<A extends Aggregate, R>(aggregateOrId: A | string, callback: (a: A) => Promise<R>): Promise<R> {
        const aggregate = await this.getAggregate(aggregateOrId);
        if (aggregate.events.length > 0) {
            Logger.warn(`${aggregate._type} ${aggregate.id} has ${aggregate.events.length} events!!`);
        }
        const hashBefore = aggregate.hash;

        const result = await callback(aggregate);

        if (aggregate.hash !== hashBefore) {
            this.client.updateDocument(aggregate);
            await this.publishEvents(aggregate);
        } else {
            Logger.debug(`${aggregate._type} ${aggregate.id} didn't change, ${aggregate.events.length} events skipped`);
            aggregate.clearEvents();
        }

        return result;
    }

    public static async delete<A extends Aggregate, R>(aggregateOrId: A | string, callback: (a: A) => Promise<R>): Promise<R> {
        const aggregate = await this.getAggregate(aggregateOrId);

        const result = await callback(aggregate);

        this.client.deleteDocument(aggregate);
        await this.publishEvents(aggregate);

        return result;
    }

    private static get client(): FirestoreClient {
        return Context.resolve(FirestoreClient);
    }

    private static async publishEvents<A extends Aggregate>(aggregate: A): Promise<void> {
        const events = aggregate.events;
        aggregate.clearEvents();

        for (const event of events) {
            await Context.bus.publish(event, ExecutionMode.Synchronous);
        }
    }

    private static async getAggregate<A extends Aggregate>(aggregateOrId: A | string): Promise<A> {
        if (typeof aggregateOrId === 'string' || aggregateOrId instanceof String) {
            return this.client._find<A>(`/aggregates/${aggregateOrId}`);
        }
        return aggregateOrId;
    }
}
