import 'reflect-metadata';

import { Command, Event, guid } from '@my-devkit/core';
import { ExecutionMode } from './execution-mode-enum';

import { Type } from './interfaces';
import { RequestMethod } from './request-method.enum';
import { Server } from './server';

/* eslint-disable @typescript-eslint/ban-types */
class ReflectHelper<T extends ReflectTarget<any>> {
    private readonly HANDLER_META_DATA = '__handler__';
    private readonly CONTROLLER_META_DATA = '__controller__';
    private readonly INJECTABLE_META_DATA = '__injectable__';
    private readonly TRANSACTIONAL_CLIENT_META_DATA = '__transactional_client__';

    constructor(private target: T, private methodName: string | symbol) {

    }

    public getParameters(): Type[] {
        return Reflect.getMetadata('design:paramtypes', this.target, this.methodName) || [];
    }

    public registerInjectable(): void {
        this.defineMetaData(this.INJECTABLE_META_DATA, guid());
    }

    public getInjectableId(): string {
        return this.getMetaData<string>(this.INJECTABLE_META_DATA)
    }

    public registerTransactionalClient(): void {
        this.defineMetaData(this.TRANSACTIONAL_CLIENT_META_DATA, guid);
    }

    public isTransactionalClient(): boolean {
        return this.hasMetaData(this.TRANSACTIONAL_CLIENT_META_DATA)
    }

    public getControllerConfiguration(): ControllerConfiguration {
        if (!this.hasMetaData(this.CONTROLLER_META_DATA)) {
            this.defineMetaData<ControllerConfiguration>(this.CONTROLLER_META_DATA, {
                basePath: null,
                routes: []
            });
        }

        return this.getMetaData<ControllerConfiguration>(this.CONTROLLER_META_DATA);
    }

    public registerController(basePath: string): void {
        this.getControllerConfiguration().basePath = basePath;
    }

    public registerControllerRoute(route: { path: string, requestMethod: RequestMethod }): void {
        this.getControllerRoute().path = route.path;
        this.getControllerRoute().requestMethod = route.requestMethod;
    }

    public registerControllerRouteArgumentInjector<T>(injector: RequestArgumentInjector<T>, index: number): void {
        this.getControllerRoute().argumentInjectors[index] = injector;
    }

    public getHandlerConfiguration(): HandlerConfiguration {
        if (!this.hasMetaData(this.HANDLER_META_DATA)) {
            this.defineMetaData<HandlerConfiguration>(this.HANDLER_META_DATA, {
                commandHandlers: new Map<string, CommandHandler>(),
                eventHandlers: new Map<string, EventHandler>()
            });
        }

        return this.getMetaData<HandlerConfiguration>(this.HANDLER_META_DATA);
    }

    public registerCommandHandler<C extends Command>(command: new () => C): void {
        this.getHandlerConfiguration().commandHandlers.set(this.methodName.toString(), {
            methodName: this.methodName.toString(),
            command
        });
    }

    public registerEventHandler<E extends Event>(event: new () => E, executionMode: ExecutionMode): void {
        this.getHandlerConfiguration().eventHandlers.set(this.methodName.toString(), {
            methodName: this.methodName.toString(),
            executionMode,
            event
        });
    }

    private getControllerRoute(): ControllerRoute {
        const config = this.getControllerConfiguration();
        if (!config.routes.find(r => r.methodName === this.methodName.toString())) {
            config.routes.push({
                methodName: this.methodName.toString(),
                path: null,
                requestMethod: null,
                argumentInjectors: []
            });
        }

        return config.routes.find(r => r.methodName === this.methodName.toString());
    }

    private getMetaData<M>(key: string): M {
        return Reflect.getMetadata(key, this.targetConstructor);
    }

    private defineMetaData<M>(key: string, data: M): void {
        Reflect.defineMetadata(key, data, this.targetConstructor);
    }

    private hasMetaData(key: string): boolean {
        return Reflect.hasMetadata(key, this.targetConstructor);
    }

    private get targetConstructor() {
        if (this.target['prototype']) {
            return this.target['prototype'].constructor;
        }

        if (this.target.constructor) {
            return this.target.constructor;
        }

        return this.target;
    }
}

export interface ControllerConfiguration {
    basePath: string;
    routes: ControllerRoute[];
}

export interface ControllerRoute {
    path: string;
    methodName: string;
    requestMethod: RequestMethod;
    argumentInjectors: RequestArgumentInjector<any>[];
}

export interface CommandHandler {
    methodName: string;
    command: new () => Command;
}

export interface EventHandler {
    methodName: string;
    executionMode: ExecutionMode;
    event: new () => Event;
}

export interface HandlerConfiguration {
    commandHandlers: Map<string, CommandHandler>;
    eventHandlers: Map<string, EventHandler>;
}

export type RequestArgumentInjector<T = any> = (req: Server.Request) => T;

export function reflect<T extends ReflectTarget<any>>(target: T, methodName?: string | symbol): ReflectHelper<T> {
    return new ReflectHelper(target, methodName);
}

type ReflectTarget<T> = T | (new (...args: any[]) => T);
