import cors from 'cors';
import express from 'express';
import { auth, firestore } from 'firebase-admin';
import { StatusCodes } from 'http-status-codes';

import { Bus } from '../bus';
import { RequestMethod } from '../request-method.enum';
import * as middlewares from './middlewares';

function encapsulate(handler: Server.RequestHandler): Server.RequestHandler {
    return async (request: Server.Request, response: Server.Response, next: Server.NextFunction): Promise<void> => {
        try {
            await handler(request, response, next);
        } catch (error) {
            next(error);
        }
    };
}

export class Server {
    private _app: express.Express;
    public static statusCodes = StatusCodes;

    constructor(options: Server.Options) {
        this._app = express();
        this.registerRequestHandler(cors({ origin: true }));
        this.registerRequestHandler(middlewares.CreateContextMiddleware(options.bus));
        this.registerRequestHandler(middlewares.HeadersMiddleWare());
        this.registerRequestHandler(middlewares.AuthenticationMiddleware(options.firebaseAuth, options.publicPaths));
    }

    public registerRequestHandler(handler: Server.RequestHandler): void {
        this._app.use(encapsulate(handler));
    }

    public registerErrorHandler(handler: Server.ErrorHandler): void {
        this._app.use(async (error: any, request: Server.Request, response: Server.Response, next: Server.NextFunction): Promise<void> => {
            try {
                await handler(error, request, response, next);
            } catch (error) {
                next(error);
            }
        });
    }

    public registerRouter(path: string, router: express.Router): void {
        this._app.use(path, router);
    }

    public get app(): express.Express {
        this.registerRequestHandler(middlewares.NotFoundMiddleware());
        this.registerErrorHandler(middlewares.ErrorMiddleWare());

        return this._app;
    }
}

export namespace Server {

    export interface Options {
        bus: Bus;
        firebaseAuth: auth.Auth;
        firestore: firestore.Firestore;
        publicPaths: string[];
    }

    export class Router {
        private router: express.Router;
        constructor() {
            this.router = express.Router();
        }

        public registerRoute(method: RequestMethod, path: string, handler: RequestHandler): void {
            switch (method) {
                case RequestMethod.GET:
                    this.router.get(path, encapsulate(handler));
                    break;
                case RequestMethod.POST:
                    this.router.post(path, encapsulate(handler));
                    break;
                default:
                    throw new Error(`Server.Router:registerRoute: ${method} is not managed yet...`);
            }
        }

        public registerOnServer(path: string, server: Server): void {
            server.registerRouter(path, this.router);
        }
    }

    export type Request = express.Request;
    export type Response = express.Response;
    export type NextFunction = express.NextFunction;
    export type RequestHandler = (request: Request, response: Response, next: NextFunction) => Promise<void> | void;
    export type ErrorHandler = (error: any, request: Request, response: Response, next: NextFunction) => Promise<void> | void;
}
